#include "posprinter.h"

#include <marketplace.h>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <algorithm>
#include <boost/algorithm/string.hpp>
#include <exception>
#include <iostream>
#include <sstream>
#include <string>

const std::string PosPrinter::Command::RESET = {0x1b, 0x40};
const std::string PosPrinter::Command::ENCODING = {'\x1b', '\x74', 16};
const std::string PosPrinter::Command::CENTERING = {'\x1b', '\x61', '\x01'};
const std::string PosPrinter::Command::LEFT_ALIGN = {0x1b, 0x61, 0x00};
const std::string PosPrinter::Command::RIGHT_ALIGN = {0x1b, 0x61, 0x02};
const std::string PosPrinter::Command::FONT_SIZE_BIG = {'\x1b', '\x21', '\x10'};
const std::string PosPrinter::Command::FONT_SIZE_NORMAL = {'\x1b', '\x21', '\x00'};
const std::string PosPrinter::Command::FEED = {0x1b, 0x64, 0x03};

PosPrinter::PosPrinter()
{
    int retValue;

    retValue = libusb_init(&contextPtr_);
    if (retValue < 0) {
        throw std::runtime_error("Init error");
    }

    libusb_device** devList;
    int devCount = libusb_get_device_list(contextPtr_, &devList);
    if (devCount <= 0) {
        libusb_exit(contextPtr_);
        throw std::runtime_error("Could not receive device list");
    }
    int numDevice = -1;
    for (int i = 0; i < devCount; ++i) {
        libusb_device_descriptor desc;
        libusb_get_device_descriptor(devList[i], &desc);
        for (const auto& supported : supportedPrinters_.models) {
            if (desc.idVendor == supported.first && desc.idProduct == supported.second) {
                numDevice = i;
                break;
            }
        }
    }

    if (numDevice < 0) {
        return;
    }

    retValue = libusb_open(devList[numDevice], &devicePtr_); // these are vendorID and productID
    if (retValue != 0) {
        libusb_free_device_list(devList, 1);
        libusb_exit(contextPtr_);
        throw std::runtime_error("Cannot open printer device");
    }

    if (libusb_kernel_driver_active(devicePtr_, 0) == 1) { // find out if kernel driver is attached
        std::cout << "Kernel driver active" << std::endl;
        if (libusb_detach_kernel_driver(devicePtr_, 0) == 0) // detach it
            std::cout << "Kernel driver detached!" << std::endl;
    }

    retValue = libusb_claim_interface(
        devicePtr_, 0); // claim interface 0 (the first) of device (mine had jsut 1)
    if (retValue < 0) {
        std::cout << "Cannot claim printer interface" << std::endl;
        throw std::runtime_error("Cannot claim printer interface");
    }

    libusb_free_device_list(devList, 1);
}

PosPrinter::~PosPrinter()
{
    if (devicePtr_) {
        int retValue;
        retValue = libusb_release_interface(devicePtr_, 0); // release the claimed interface
        if (retValue != 0) {
            std::cout << "Cannot release printer interface" << std::endl;
        } else {
            std::cout << "Printer interface released" << std::endl;
        }

        libusb_close(devicePtr_); // close the device we opened
    }
    if (contextPtr_) {
        libusb_exit(contextPtr_); // close the session
    }
}

void PosPrinter::write(const std::string& text)
{
    if (devicePtr_ == NULL)
        return;

    using namespace std::string_literals;
    std::string text_escaped = text;
    boost::replace_all(text_escaped, "ä", "\xe4");
    boost::replace_all(text_escaped, "ö", "\xf6");
    boost::replace_all(text_escaped, "ü", "\xfc");
    boost::replace_all(text_escaped, "Ä", "\xc4");
    boost::replace_all(text_escaped, "Ö", "\xd6");
    boost::replace_all(text_escaped, "Ü", "\xdc");
    boost::replace_all(text_escaped, "€", "\x80");

    int length = text_escaped.length();
    int actual{0};
    int retValue =
        libusb_bulk_transfer(devicePtr_, (0x03 | LIBUSB_ENDPOINT_OUT),
                             (unsigned char*)text_escaped.c_str(), length, &actual, 10000);
    if (retValue != 0 || actual != length)
        std::cerr << "Write Error" << std::endl;
}

void PosPrinter::printHeader()
{
    std::stringstream commandStream;

    commandStream << Command::RESET << Command::ENCODING << Command::CENTERING
                  << Command::FONT_SIZE_BIG;
    commandStream << "Kindersachenmarkt\nDettingen\n\n";
    commandStream << Command::LEFT_ALIGN << Command::Command::FONT_SIZE_NORMAL;

    write(commandStream.str());
}

void PosPrinter::printTest()
{
    std::stringstream commandStream;
    commandStream << Command::ENCODING;
    commandStream << "Der Drucker kann von KIMA2\nangesprochen werden.\n\n"
                  << "Beachten Sie, dass nicht\nalle Modelle Strichcodes\nausdrucken können.";
    commandStream << Command::FEED;
    printHeader();
    write(commandStream.str());
}

void PosPrinter::printSaleReceipt(Sale* sale)
{
    std::stringstream commandStream;
    printHeader();
    commandStream << Command::RESET << Command::ENCODING << Command::RIGHT_ALIGN;
    commandStream << sale->getTimestamp() << "\n\n";
    commandStream << Command::LEFT_ALIGN;
    for (const auto& article : sale->getArticles()) {
        commandStream << "Art. " << article->getCompleteArticleNo() << "........... "
                      << article->getPriceAsString() << "\n";
    }
    commandStream << "\nGesamt................. " << sale->sumAsString() << "\n";
    commandStream << Command::CENTERING;
    commandStream << "\n\nVielen Dank für Ihren Einkauf!";
    commandStream << Command::LEFT_ALIGN << Command::FEED;
    write(commandStream.str());
}

void PosPrinter::printSellerReceipt(Seller* seller, int percent, int maxFeeInCent)
{
    std::stringstream commandStream;
    printHeader();
    commandStream << Command::RESET << Command::ENCODING << Command::RIGHT_ALIGN;
    std::string timeStr =
        boost::posix_time::to_simple_string(boost::posix_time::second_clock::local_time());
    commandStream << timeStr << "\n\n";
    commandStream << Command::LEFT_ALIGN;
    commandStream << "Verkäufernummer: " << std::setw(4) << seller->getSellerNo() << "\n\n";
    commandStream << "Anzahl Artikel geliefert: " << std::setw(4) << seller->numArticlesOffered()
                  << "\n";
    commandStream << "Anzahl Artikel verkauft:  " << std::setw(4) << seller->numArticlesSold()
                  << "\n\n";
    if (seller->getArticles(true).size() == 0) {
        commandStream << "Verkaufte Artikel:\n        *** keine ***\n";
    } else {
        commandStream << "Verkaufte Artikel:\n";
    }
    for (auto& article : seller->getArticles(true)) {
        commandStream << "Art. " << article->getCompleteArticleNo() << "........... "
                      << article->getPriceAsString() << "\n";
    }
    commandStream << "\nGesamt................." << seller->sumAsString() << "\n";
    commandStream << "./. Gebühr............."
                  << marketFeeAsString(seller->sumInCents(), percent, maxFeeInCent) << "\n";
    commandStream << "\nAuszahlung............."
                  << paymentAsString(seller->sumInCents(), percent, maxFeeInCent) << "\n";
    commandStream << Command::FEED;

    write(commandStream.str());
}

bool PosPrinter::isValid()
{
    if (devicePtr_)
        return true;
    else
        return false;
}